// NtInjectUtil.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include "NtInjectUtil.h"
#include "..\psapi\psapi.h"

HINSTANCE g_hInst;

BOOL SetPrivilege( HANDLE hToken, LPCTSTR Privilege,  BOOL bEnablePrivilege );

// LoadLibraryRemote
//
// Works just like LoadLibrary, except on a specified target process.
//
HMODULE APIENTRY LoadLibraryRemote( HANDLE hProcess, wchar_t *pszLibFileName )
{
    HMODULE hLib = 0;
    DWORD   cbLibName = (wcslen(pszLibFileName) * sizeof(wchar_t)) + 1;

    // Allocate memory in the target process that will hold the
    // name of the DLL to load.
    //
    void *pvRemMem = VirtualAllocEx( hProcess, 0, cbLibName,
                                     MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );

    if( pvRemMem )
    {
        // Copy the DLL name to load into the target process.
        //
        if( WriteProcessMemory(hProcess, pvRemMem, pszLibFileName, cbLibName, 0) )
        {
            HANDLE hRemThread;

            // Start a new thread in the target process.  The ThreadProc will
            // be the address of LoadLibrary.  The argument to the "ThreadProc"
            // will be the address of the DLL name to load.  Effectively, this
            // will execute LoadLibrary(pszLibName) in the target process, then
            // immediately exit the new thread.
            //
            hRemThread = CreateRemoteThread( hProcess, 0, 0,
                                             (LPTHREAD_START_ROUTINE)LoadLibraryW,
                                             pvRemMem,
                                             0, 0);

            if( hRemThread )
            {
                WaitForSingleObject(hRemThread, INFINITE);
                GetExitCodeThread(hRemThread, (LPDWORD)&hLib);

                CloseHandle(hRemThread);
            }
        }

        VirtualFreeEx(hProcess, pvRemMem, 0, MEM_RELEASE);
    }

    return(hLib);
}

// FreeLibraryRemote
//
// Undoes the work performed by LoadLibraryRemote.
//
BOOL APIENTRY FreeLibraryRemote( HANDLE hProcess, HMODULE hLibModule )
{
    HANDLE  hRemThread;
    DWORD   dwExitCode = 0;

    // Start a new thread in the target process.  The ThreadProc will
    // be the address of LoadLibrary.  The argument to the "ThreadProc"
    // will be the address of the DLL name to load.  Effectively, this
    // will execute LoadLibrary(pszLibName) in the target process, then
    // immediately exit the new thread.
    //
    hRemThread = CreateRemoteThread( hProcess, 0, 0,
                                     (LPTHREAD_START_ROUTINE)FreeLibrary,
                                     (LPVOID)hLibModule,
                                     0, 0 );

    if( hRemThread )
    {
        WaitForSingleObject(hRemThread, INFINITE);
        GetExitCodeThread(hRemThread, &dwExitCode);
        CloseHandle(hRemThread);
    }

    return BOOL(dwExitCode != 0);
}

// CreateRemoteThreadEx
//
// Creates a thread in a target process using a named {dll, procname} tuple
// to identify the thread proc address.  The specified DLL is injected into the
// target process, then the thread is started.  The thread handle, id, and
// remote library handle are all returned.
//
// If this function returns TRUE, the caller is responsible for cleaning
// up the remote library and thread:
//
//      o Use WaitForXxxObject(*phThread) to wait for the remote thread to exit.
//      o Use CloseHandle(*phThread) to close the remote thread handle.
//      o Use FreeLibraryRemote(*phLib) to unload the dll that was injected.
//
BOOL APIENTRY CreateRemoteThreadEx( // IN:
                                    HANDLE hProcess, LPSECURITY_ATTRIBUTES lpThreadAttributes,
                                    DWORD dwStackSize, wchar_t *pszLibName, char *pszThreadProcName,
                                    LPVOID lpParameter, DWORD dwCreationFlags,
                                    
                                    // OUT:
                                    HANDLE *phThread, LPDWORD lpThreadId, HMODULE *phLib )
{
    BOOL    fOutcome = FALSE;
    HMODULE hModule;
    DWORD   cbThreadProc = 0;

    // Calculate the offset of the thread proc relative to the HMODULE.
    //
    hModule = LoadLibraryExW(pszLibName, 0, DONT_RESOLVE_DLL_REFERENCES);

    if( hModule )
    {
        DWORD pfxnThreadProc = (DWORD)GetProcAddress(hModule, pszThreadProcName);

        if( pfxnThreadProc )
        {
            cbThreadProc = pfxnThreadProc - (DWORD)hModule;
            fOutcome = TRUE;
        }

        FreeLibrary(hModule);
    }

    // Inject the DLL into the target process, then spin up a thread that
    // calls into that DLL.
    //
    if( fOutcome  )
    {
        fOutcome = FALSE;

        hModule = LoadLibraryRemote(hProcess, pszLibName);

        if( hModule )
        {
            HANDLE hThread;
            
            hThread = CreateRemoteThread( hProcess, lpThreadAttributes, dwStackSize,
                                          (LPTHREAD_START_ROUTINE)((DWORD)hModule + cbThreadProc),
                                          lpParameter, dwCreationFlags, lpThreadId );
            if( hThread )
            {
                fOutcome = TRUE;

                *phThread = hThread;
                *phLib = hModule;
            }
            else
            {
                FreeLibraryRemote(hProcess, hModule);
            }
        }
    }

    return(fOutcome);
}

// AllocateRemoteMemory
//
// Allocates memory in a target process address space, then copies the
// contents of the buffer specified by pv over to the target address
// space.  The address of the newly allocated buffer in the target
// process is returned (NULL on failure).
//
void * APIENTRY AllocateRemoteMemory( HANDLE hProcess, void *pv, DWORD cb )
{
    // Allocate memory in the target process.
    //
    void *pvRemMem = VirtualAllocEx( hProcess, 0, cb,
                                     MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );

    if( pvRemMem )
    {
        // Copy contents of local copy buffer over to target process.
        //
        if( !WriteProcessMemory(hProcess, pvRemMem, pv, cb, 0) )
        {
            FreeRemoteMemory(hProcess, pvRemMem);
            pvRemMem = 0;
        }
    }

    return(pvRemMem);
}

// FreeRemoteMemory
//
// Undoes the work performed by AllocateRemoteMemory.
//
void APIENTRY FreeRemoteMemory( HANDLE hProcess, void *pv )
{
    VirtualFreeEx(hProcess, pv, 0, MEM_RELEASE);
}

// OpenProcessByName
//
// Works like OpenProcess, but takes a string that identifies
// the process to open instead of a process ID.  The string
// represents that base filename.ext form of the target process.
//
HANDLE APIENTRY OpenProcessByName( DWORD dwDesiredAccess, BOOL bInheritHandle, wchar_t *pszName )
{
    enum { MAX_PROCESSES = 128 };

    HANDLE  hTargetProcess = 0;
    DWORD   rgdwPID[MAX_PROCESSES];
    DWORD   cb = 0;

    if( EnumProcesses(rgdwPID, sizeof(rgdwPID), &cb) )
    {
        DWORD cProcesses = cb / sizeof(DWORD);

        for( DWORD n = 0; (0 == hTargetProcess) && (n < cProcesses); n++ )
        {
            HANDLE hProcess = OpenProcess(dwDesiredAccess, bInheritHandle, rgdwPID[n]);

            if( hProcess )
            {
                HMODULE hModule;

                if( EnumProcessModules(hProcess, &hModule, sizeof(hModule), &cb) )
                {
                    wchar_t szBaseName[MAX_PATH+1];

                    if( GetModuleBaseNameW(hProcess, hModule, szBaseName, MAX_PATH) &&
                        !wcsicmp(szBaseName, pszName) )
                    {
                        // Found it.
                        hTargetProcess = hProcess;
                    }
                }

                if( hTargetProcess != hProcess )
                {
                    // Only close the process handle if this isn't the
                    // one we're looking for.
                    //
                    CloseHandle(hProcess);
                }
            }
        }
    }

    return(hTargetProcess);
}

BOOL SetPrivilege( HANDLE hToken, wchar_t *Privilege,  BOOL bEnablePrivilege )
{ 
    TOKEN_PRIVILEGES tp;
    LUID luid;
    TOKEN_PRIVILEGES tpPrevious;
    DWORD cbPrevious=sizeof(TOKEN_PRIVILEGES);

    if( !LookupPrivilegeValue( NULL, Privilege, &luid ) )
    {
        return FALSE;
    }

    // Get current privilege setting.
    //
    tp.PrivilegeCount           = 1;
    tp.Privileges[0].Luid       = luid;
    tp.Privileges[0].Attributes = 0;

    AdjustTokenPrivileges( hToken, FALSE, &tp, sizeof(TOKEN_PRIVILEGES),
                           &tpPrevious, &cbPrevious );

    if( GetLastError() != ERROR_SUCCESS )
    {
        return FALSE;
    }

    // Set privilege based on previous setting.
    //
    tpPrevious.PrivilegeCount       = 1;
    tpPrevious.Privileges[0].Luid   = luid;

    if( bEnablePrivilege )
    {
        tpPrevious.Privileges[0].Attributes |= (SE_PRIVILEGE_ENABLED);
    }
    else
    {
        tpPrevious.Privileges[0].Attributes ^=
            (SE_PRIVILEGE_ENABLED & tpPrevious.Privileges[0].Attributes);
    }

    AdjustTokenPrivileges( hToken, FALSE, &tpPrevious, cbPrevious,
                           NULL, NULL );

    if( GetLastError() != ERROR_SUCCESS )
    {
        return FALSE;
    }

    return TRUE;
} 

BOOL APIENTRY EnableDebugPrivilege( void )
{
    HANDLE  hToken;
    BOOL    fOkay;

    fOkay = OpenProcessToken( GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY,
                              &hToken );

    if( fOkay )
    {
        fOkay = SetPrivilege(hToken, SE_DEBUG_NAME, TRUE);
        CloseHandle(hToken);
    }

    return(fOkay);
}

BOOL APIENTRY DllMain( HINSTANCE hInst, DWORD dwReason, LPVOID pvReserved ) 
{
    switch( dwReason )
	{
		case DLL_PROCESS_ATTACH:
            g_hInst= hInst;
        break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
        break;
    }

    return(TRUE);
}
