// cmdline.exe
//
// Utility to display the command line of any running process on NT.
//
// Mike Woodring
// http://www.bearcanyon.com
//
#include "stdafx.h"
#include "cmdlinespy\cmdlinespy.h"

#ifdef _UNICODE
#define _tcout wcout
#else
#define _tcout cout
#endif

void GetSpyLibName( wchar_t *pszLib, DWORD cch )
{
    const wchar_t *pszBaseName = L"cmdlinespy.dll";
    const DWORD BASENAMELEN = wcslen(pszBaseName);

    GetModuleFileNameW(0, pszLib, cch - BASENAMELEN - 1);

    wchar_t *pszSlash = wcsrchr(pszLib, L'\\');
    wcscpy(pszSlash+1, pszBaseName);
}

TCHAR *GetProcessCmdLine( DWORD dwPID )
{
    TCHAR *pszCmdLine = 0;

    HANDLE hProcess = OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwPID);

    if( hProcess )
    {
        HANDLE  hRemThread = 0;
        HMODULE hLibCmdLineSpy = 0;
        wchar_t szSpyLib[MAX_PATH + 1];

        GetSpyLibName(szSpyLib, MAX_PATH);

        if( CreateRemoteThreadEx( hProcess, 0, 0,
                                  szSpyLib,
#ifdef _UNICODE
                                  "GetCmdLineExW",
#else
                                  "GetCmdLineExA",
#endif
                                  0, 0,
                                  &hRemThread, 0, &hLibCmdLineSpy ) )
        {
            COMMANDLINEINFO *pRemCmdLineInfo = 0;

            WaitForSingleObject(hRemThread, INFINITE);
            GetExitCodeThread(hRemThread, (LPDWORD)&pRemCmdLineInfo);

            if( pRemCmdLineInfo )
            {
                COMMANDLINEINFO CmdLineInfo;

                if( ReadProcessMemory( hProcess, pRemCmdLineInfo,
                                       &CmdLineInfo, sizeof(CmdLineInfo), 0 ) )
                {
                    DWORD cchWithNul = CmdLineInfo.cchPath + 1;

                    pszCmdLine = new TCHAR[cchWithNul];

                    if( !ReadProcessMemory( hProcess, CmdLineInfo.pszCmdLine,
                                            pszCmdLine, cchWithNul * sizeof(TCHAR), 0 ) )
                    {
                        delete [] pszCmdLine;
                        pszCmdLine = 0;
                    }
                }

                FreeRemoteMemory(hProcess, pRemCmdLineInfo);
            }

            CloseHandle(hRemThread);
        }

        if( hLibCmdLineSpy )
        {
            FreeLibraryRemote(hProcess, hLibCmdLineSpy);
        }

        CloseHandle(hProcess);
    }

    return(pszCmdLine);
}

int _tmain( int cArgs, TCHAR *pszArgs[] )
{
    if( cArgs == 2 )
    {
        EnableDebugPrivilege();

        DWORD dwPID = _ttol(pszArgs[1]);
        TCHAR *pszCmdLine = GetProcessCmdLine(dwPID);

        if( pszCmdLine )
        {
            _tcout << pszCmdLine << endl;
            delete [] pszCmdLine;
        }
        else
        {
            _tcout << __TEXT("Could not retrieve command line for process ") << dwPID << endl;
        }
    }
    else
    {
        _tcout << __TEXT("Usage: cmdline ProcessId") << endl;
    }

	return 0;
}
