// dbgpack - MTS/COM+ package debugging utility.
//
// Mike Woodring
// http://www.bearcanyon.com
//
#include "stdafx.h"

#ifdef _UNICODE
#define _tcout wcout
#else
#define _tcout cout
#endif

void GetSpyLibName( wchar_t *pszLib, DWORD cch )
{
    const wchar_t *pszBaseName = L"cmdlinespy.dll";
    const DWORD BASENAMELEN = wcslen(pszBaseName);

    GetModuleFileNameW(0, pszLib, cch - BASENAMELEN - 1);

    wchar_t *pszSlash = wcsrchr(pszLib, L'\\');
    wcscpy(pszSlash+1, pszBaseName);
}

TCHAR *GetProcessCmdLine( DWORD dwPID )
{
    TCHAR *pszCmdLine = 0;

    HANDLE hProcess = OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwPID);

    if( hProcess )
    {
        HANDLE  hRemThread = 0;
        HMODULE hLibCmdLineSpy = 0;
        wchar_t szSpyLib[MAX_PATH + 1];

        GetSpyLibName(szSpyLib, MAX_PATH);

        if( CreateRemoteThreadEx( hProcess, 0, 0,
                                  szSpyLib,
#ifdef _UNICODE
                                  "GetCmdLineW",
#else
                                  "GetCmdLineA",
#endif
                                  0, 0,
                                  &hRemThread, 0, &hLibCmdLineSpy ) )
        {
            TCHAR *pszRemCmdLine = 0;

            WaitForSingleObject(hRemThread, INFINITE);
            GetExitCodeThread(hRemThread, (LPDWORD)&pszRemCmdLine);

            if( pszRemCmdLine )
            {
                pszCmdLine = new TCHAR[256];
                ZeroMemory(pszCmdLine, 256);

                if( !ReadProcessMemory( hProcess, pszRemCmdLine,
                                        pszCmdLine, 255 * sizeof(TCHAR), 0 ) )
                {
                    delete [] pszCmdLine;
                    pszCmdLine = 0;
                }
            }

            CloseHandle(hRemThread);
        }

        if( hLibCmdLineSpy )
        {
            FreeLibraryRemote(hProcess, hLibCmdLineSpy);
        }

        CloseHandle(hProcess);
    }

    return(pszCmdLine);
}

typedef list<DWORD> PROCESSIDLIST;

BOOL RunningOnWin2k( void )
{
    BOOL fOnWin2k = FALSE;
    OSVERSIONINFO VerInfo;

    ZeroMemory(&VerInfo, sizeof(VerInfo));
    VerInfo.dwOSVersionInfoSize = sizeof(VerInfo);
    
    if( GetVersionEx(&VerInfo) )
    {
        fOnWin2k = BOOL(VerInfo.dwMajorVersion >= 5);
    }

    return(fOnWin2k);
}

void EnumActiveSurrogateProcesses( PROCESSIDLIST& listServers )
{
    enum { MAX_PROCESSES = 128 };

    DWORD   rgdwPID[MAX_PROCESSES];
    DWORD   cb = 0;

    if( EnumProcesses(rgdwPID, sizeof(rgdwPID), &cb) )
    {
        DWORD cProcesses = cb / sizeof(DWORD);

        for( DWORD n = 0; n < cProcesses; n++ )
        {
            HANDLE hProcess = OpenProcess(PROCESS_ALL_ACCESS, FALSE, rgdwPID[n]);

            if( hProcess )
            {
                HMODULE hModule;

                if( EnumProcessModules(hProcess, &hModule, sizeof(hModule), &cb) )
                {
                    TCHAR szBaseName[MAX_PATH+1];
                    TCHAR szTargetExeName[MAX_PATH+1];

                    if( RunningOnWin2k() )
                    {
                        _tcscpy(szTargetExeName, __TEXT("dllhost.exe"));
                    }
                    else
                    {
                        _tcscpy(szTargetExeName, __TEXT("mtx.exe"));
                    }

                    if( GetModuleBaseName(hProcess, hModule, szBaseName, MAX_PATH) &&
                        !_tcsicmp(szBaseName, szTargetExeName) )
                    {
                        listServers.push_back(rgdwPID[n]);
                    }
                }

                CloseHandle(hProcess);
            }
        }
    }
}

void GetJITDebuggerStartString( TCHAR *pszFormat, DWORD cch )
{
    LONG    lResult;
    HKEY    hkeySubkey;

    lResult =
        RegOpenKeyEx( HKEY_LOCAL_MACHINE,
                      __TEXT("Software\\Microsoft\\Windows NT\\CurrentVersion\\AeDebug"),
                       0, KEY_ALL_ACCESS, &hkeySubkey );

    if( ERROR_SUCCESS == lResult )
    {
        DWORD dwType = REG_SZ;
        DWORD cb = (cch * sizeof(TCHAR));

        lResult =
            RegQueryValueEx( hkeySubkey, __TEXT("Debugger"),
                             0, &dwType, (LPBYTE)pszFormat, &cb );

        RegCloseKey(hkeySubkey);
    }

    if( ERROR_SUCCESS != lResult )
    {
        // We'll only take this path if a JIT debugger hasn't
        // been set, or this utility is running in the context
        // of LOCAL SYSTEM, where HKCU doesn't have any meaning.
        //
        _tcscpy(pszFormat, __TEXT("msdev.exe -p %ld"));
    }
}

BOOL AttachDebuggerToProcess( DWORD dwPID )
{
    STARTUPINFO         si = { sizeof(si), 0, __TEXT("") };
    PROCESS_INFORMATION pi = {0};
    BOOL                fOkay;
    TCHAR               szDbgCmdLineFormat[128];
    TCHAR               szCmdLine[128];

    GetJITDebuggerStartString(szDbgCmdLineFormat, 128);
    _stprintf(szCmdLine, szDbgCmdLineFormat, dwPID, dwPID);

#if 1
    fOkay = CreateProcess( 0, szCmdLine, 0, 0, FALSE, 0,
                           0, 0, &si, &pi );

    if( fOkay )
    {
        CloseHandle(pi.hProcess);
        CloseHandle(pi.hThread);
    }
#else
    // Trying to find a way to start a new copy of VC if one's
    // not already running, or reuse an existing VC if one's available.
    //
    TCHAR *pszExePath = szCmdLine;
    TCHAR *pszDebuggerArgs;

    if( pszExePath[0] == __TEXT('\"') )
    {
        // Skip leading " character.
        //
        pszExePath += 1;

        // Find and lop off trailing " character.
        //
        pszExePath = _tcschr(pszExePath, __TEXT('\"'));
        *pszExePath = 0;

        // Note where the arguments are, then rewind
        // to the beginning.
        //
        pszDebuggerArgs = pszExePath + 1;
        pszExePath = &szCmdLine[1];
    }
    else
    {
        // Path to debugger exe does not contain surrounding
        // " characters.
        //
        pszDebuggerArgs = _tcsstr(szCmdLine, __TEXT(".exe"));
        pszDebuggerArgs += _tcslen(__TEXT(".exe"));

        *pszDebuggerArgs = 0;
        pszDebuggerArgs += 1;
    }


    if( ShellExecute( HWND_DESKTOP, __TEXT("open"), pszExePath,
                      pszDebuggerArgs, 0, SW_SHOWNORMAL ) > (HINSTANCE)32 )
    {
        fOkay = TRUE;
    }
    else
    {
        fOkay = FALSE;
    }
#endif

    return(fOkay);
}

BOOL StartDebuggerForPackage( TCHAR *pszAppId )
{
    STARTUPINFO         si = { sizeof(si), 0, __TEXT("") };
    PROCESS_INFORMATION pi = {0};
    BOOL                fOkay;
    TCHAR              *szCmdLine = new TCHAR[(MAX_PATH * 2) + 1];

    GetJITDebuggerStartString(szCmdLine, 128);

    TCHAR *pszExe = __TEXT(".exe");
    TCHAR *pszExt = _tcsstr(szCmdLine, pszExe);

    pszExt += _tcslen(pszExe);
    if( pszExt[0] == __TEXT('\"') )
    {
        pszExt += 1;
    }
    *pszExt = 0;

    int cchLen = ((DWORD)pszExt - (DWORD)szCmdLine) / sizeof(TCHAR);

    TCHAR *pszSurrogateName = (RunningOnWin2k() ? __TEXT("dllhost.exe") : __TEXT("mtx.exe"));

    TCHAR szSystemDir[MAX_PATH + 1];
    GetSystemDirectory(szSystemDir, MAX_PATH);

    _stprintf( &szCmdLine[cchLen], __TEXT(" %s\\%s %s:%s"),
               szSystemDir,
               pszSurrogateName,
               (RunningOnWin2k() ? __TEXT("/ProcessId") : __TEXT("/p")),
               pszAppId );

    fOkay = CreateProcess( 0, szCmdLine, 0, 0, FALSE, 0,
                           0, 0, &si, &pi );

    if( fOkay )
    {
        CloseHandle(pi.hProcess);
        CloseHandle(pi.hThread);
    }

    delete [] szCmdLine;
    return(fOkay);
}

int _tmain( int cArgs, TCHAR *pszArgs[] )
{
    EnableDebugPrivilege();

    PROCESSIDLIST listServers;

    // Enumerate running MTS/COM+ server packages
    // process surrogates (mtx.exe on NT4, dllhost.exe on COM+).
    //
    EnumActiveSurrogateProcesses(listServers);

    PROCESSIDLIST::iterator it;
    PROCESSIDLIST::iterator begin = listServers.begin();
    PROCESSIDLIST::iterator end = listServers.end();

    BOOL    fAttach = FALSE;
    TCHAR  *pszAppId = 0;

    if( (cArgs >= 2) && (0 == _tcsnicmp(pszArgs[1], __TEXT("/p:"), 3)) )
    {
        fAttach = TRUE;
        pszAppId = &pszArgs[1][3];
    }

    BOOL fFound = FALSE;
    DWORD dwPID = 0;

    for( it = begin; !fFound && (it != end); ++it )
    {
        TCHAR *pszCmdLine = GetProcessCmdLine(*it);

        if( pszCmdLine )
        {
#ifdef _DEBUG
            _tcout << endl;
            _tcout << __TEXT("Surrogate process ID ") << *it << __TEXT(": ") << endl;
            _tcout << __TEXT(" ") << pszCmdLine << endl;
#endif

            if( fAttach && _tcsstr(pszCmdLine, pszAppId) )
            {
                fFound = TRUE;
                dwPID = *it;
            }

            delete [] pszCmdLine;
        }
    }

    if( fAttach )
    {
        if( fFound )
        {
            _tcout << __TEXT("Attaching debugger to process ID: ") << dwPID << endl;

            if( !AttachDebuggerToProcess(dwPID) )
            {
                _tcout << __TEXT("Failed to start debugger.") << endl;
            }
        }
        else
        {
            _tcout << __TEXT("Package isn't started yet.  Launching debugger.") << endl;

            // The specified package isn't already running, so start
            // VC and tell it to pre-start the package.
            //
            if( !StartDebuggerForPackage(pszAppId) )
            {
                _tcout << __TEXT("Failed to start debugger.") << endl;
            }
        }
    }

	return 0;
}
